import { cloneDeep, isEqual, isUndefined, merge, omit, omitBy } from "lodash-es";
import type { SubmitOptions } from "@/js/contexts/shared/composables/useFetch";
import useFetch from "@/js/contexts/shared/composables/useFetch";
import type { Ref } from "vue";

type FormDataType = object;
type FormDataConvertible = Array<FormDataConvertible> | { [key: string]: FormDataConvertible } | Blob | FormDataEntryValue | Date | boolean | number | null | undefined;
type RequestPayload = Record<string, FormDataConvertible> | FormData;

function hasFiles(data: RequestPayload | FormDataConvertible): boolean {
  return (
    data instanceof File ||
    data instanceof Blob ||
    (data instanceof FileList && data.length > 0) ||
    (data instanceof FormData && Array.from(data.values()).some((value) => hasFiles(value))) ||
    (typeof data === "object" && data !== null && Object.values(data).some((value) => hasFiles(value)))
  );
}

function composeKey(parent: string | null, key: string): string {
  return parent ? `${parent}[${key}]` : key;
}

function objectToFormData(source: Record<string, FormDataConvertible>, form: FormData = new FormData(), parentKey: string | null = null): FormData {
  source = source || {};
  for (const key in source) {
    if (Object.prototype.hasOwnProperty.call(source, key)) {
      append(form, composeKey(parentKey, key), source[key]);
    }
  }
  return form;
}

function append(form: FormData, key: string, value: FormDataConvertible): void {
  if (Array.isArray(value)) {
    return Array.from(value.keys()).forEach((index) => append(form, composeKey(key, index.toString()), value[index]));
  } else if (value instanceof Date) {
    return form.append(key, value.toISOString());
  } else if (value instanceof File) {
    return form.append(key, value, value.name);
  } else if (value instanceof Blob) {
    return form.append(key, value);
  } else if (typeof value === "boolean") {
    return form.append(key, value ? "1" : "0");
  } else if (typeof value === "string") {
    return form.append(key, value);
  } else if (typeof value === "number") {
    return form.append(key, `${value}`);
  } else if (value === null || value === undefined) {
    return form.append(key, "");
  }

  objectToFormData(value, form, key);
}

export default function useForm<TForm extends FormDataType>(data: TForm) {
  const isDirty = ref<boolean>(false);
  const isFetching = ref<boolean>(false);
  const defaults = ref<TForm>(cloneDeep(data));
  const values = ref<TForm>(cloneDeep(data));
  const errors = ref<Partial<Record<keyof TForm | "unknown", string[]>>>({}) as Ref<Partial<Record<keyof TForm | "unknown", string[]>>>;

  const setErrors = (e: Record<keyof TForm, string> | string) => {
    merge(errors.value, typeof e === "object" ? e : { unknown: e });
  };
  const clearErrors = () => {
    errors.value = {};
  };
  const submit = async (method: "get" | "post" | "put" | "patch" | "delete", url: string, options?: SubmitOptions & { except?: (keyof TForm)[] }) => {
    let value = cloneDeep(values.value) as RequestPayload;
    if (hasFiles(value) && !(value instanceof FormData)) {
      value = objectToFormData(value);
    }
    await useFetch(url).fetch(method, omitBy(omit(value, options?.except ?? []), isUndefined), {
      onStart: () => {
        isFetching.value = true;
        if (options?.onStart) {
          options.onStart();
        }
      },
      onSuccess: (response) => {
        isFetching.value = false;
        clearErrors();
        if (options?.onSuccess) {
          options.onSuccess(response);
        }
        defaults.value = cloneDeep(values.value);
        isDirty.value = false;
      },
      onError: (e) => {
        isFetching.value = false;
        clearErrors();
        setErrors(e.response?.data?.errors ?? e.response?.data?.message ?? e);
        if (options?.onError) {
          options.onError(e.response?.data?.errors ?? e.response?.data?.message ?? e);
        }
      },
    });
  };
  const reset = (...fields: (keyof TForm)[]) => {
    const resolvedData = cloneDeep(defaults.value);
    const clonedData = cloneDeep(resolvedData);

    if (fields.length === 0) {
      defaults.value = clonedData;
      Object.assign(values.value, resolvedData);
    } else {
      Object.keys(resolvedData)
        .filter((key) => fields.includes(key as keyof FormDataType))
        .forEach((key) => {
          defaults.value[key as keyof FormDataType] = clonedData[key as keyof FormDataType];
          values.value[key as keyof FormDataType] = resolvedData[key as keyof FormDataType];
        });
    }
  };
  const updatedValues = () => {
    defaults.value = cloneDeep(values.value);
    isDirty.value = false;
  };

  const form = reactive({
    values,
    submit,
    reset,
    clearErrors,
    updatedValues,
    errors: computed(() => errors.value),
    isDirty: computed(() => isDirty.value),
    isFetching: computed(() => isFetching.value),
    hasErrors: computed(() => Object.keys(errors.value ?? {}).length > 0),
    get: async (url: string, options?: SubmitOptions & { except?: (keyof TForm)[] }) => submit("get", url, options),
    post: async (url: string, options?: SubmitOptions & { except?: (keyof TForm)[] }) => submit("post", url, options),
    put: async (url: string, options?: SubmitOptions & { except?: (keyof TForm)[] }) => submit("put", url, options),
    patch: async (url: string, options?: SubmitOptions & { except?: (keyof TForm)[] }) => submit("patch", url, options),
    delete: async (url: string, options?: SubmitOptions & { except?: (keyof TForm)[] }) => submit("delete", url, options),
  });

  watch(form.values, () => (isDirty.value = !isEqual(values.value, defaults.value)), { immediate: true, deep: true });

  return form;
}
